#!/usr/bin/env python
#
# Copyright 2019-2020 Flavio Garcia
# Copyright 2016-2017 Veeti Paananen under MIT License
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
The account info command.
"""

from . import get_version
from .acme import AcmeV2
from .errors import AutomatoesError
import logging
import os

logger = logging.getLogger(__name__)


def info(server, account, paths):
    acme_v2 = AcmeV2(server, account)
    print("Candango Automatoes {}. Manuale replacement."
          "\n\n".format(get_version()))

    try:
        print("Requesting account data...\n")

        response = acme_v2.get_registration()
        print("  Account contacts:")
        for contact in response['contact']:
            print("    {}".format(contact[7:]))
        if "createdAt" in response:
            print("  Creation: {}".format(response['createdAt']))
        if "initialIp" in response:
            print("  Initial Ip: {}".format(response['initialIp']))
        if "key" in response:
            print("  Key Data:")
            print("    Type: {}".format(response['key']['kty']))
            print("    Public key (part I) n: {}".format(
                response['key']['n']))
            print("    Public key (part II) e: {}\n".format(
                response['key']['e']))
        else:
            print("    WARNING: Server won't return your key information.")
        print("    Private key stored at {}".format(
            os.path.join(paths['current'], "account.json")))
    except IOError as e:
        raise AutomatoesError(e)



#!/usr/bin/env python
#
# Copyright 2019-2020 Flavio Garcia
# Copyright 2016-2017 Veeti Paananen under MIT License
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json

from .crypto import (generate_jwk_thumbprint, load_private_key,
                     export_private_key)
from datetime import datetime


class Account:

    def __init__(self, key, uri=None):
        self.key = key
        self.uri = uri

    def serialize(self):
        return json.dumps({
            'key': export_private_key(self.key).decode("utf-8"),
            'uri': self.uri,
        }).encode("utf-8")

    @property
    def thumbprint(self):
        return generate_jwk_thumbprint(self.key)

    @staticmethod
    def deserialize(data):
        try:
            if not isinstance(data, str):
                data = data.decode('utf-8')
            data = json.loads(data)
            if "key" not in data or "uri" not in data:
                raise ValueError("Missing 'key' or 'uri' fields.")
            return Account(key=load_private_key(data['key'].encode("utf8")),
                           uri=data['uri'])
        except (TypeError, ValueError, AttributeError) as e:
            raise IOError("Invalid account structure: {}".format(e))


class Authorization:

    def __init__(self, contents, uri, ty_pe):
        self.contents = contents
        self.uri = uri
        self.type = ty_pe
        self.certificate_uri = None
        self.certificate = {}


class Challenge:

    def __init__(self, contents, domain, expires, status, ty_pe, key):
        self.contents = contents
        self.domain = domain
        self.expires = expires
        self.status = status
        self.type = ty_pe
        self.key = key

    def serialize(self):
        return json.dumps({
            'contents': self.contents,
            'domain': self.domain,
            'expires': self.expires,
            'status': self.status,
            'type': self.type,
            'key': self.key
        }).encode('utf-8')

    @property
    def file_name(self):
        return "{}_challenge.json".format(self.domain)


class Order:

    def __init__(self, contents, uri, ty_pe):
        self.contents = contents
        self.uri = uri
        self.type = ty_pe
        self.certificate_uri = None
        self.certificate = {}
        self.key = None

    @property
    def expired(self):
        if self.contents['status'] == "expired":
            return True
        order_timestamp = datetime.strptime(self.contents['expires'][0:19],
                                            "%Y-%m-%dT%H:%M:%S")
        return order_timestamp < datetime.now()

    @property
    def invalid(self):
        return self.contents['status'] == "invalid"

    def serialize(self):
        return json.dumps({
            'contents': self.contents,
            'uri': self.uri,
            'type': self.type,
            'certificate_uri': self.certificate_uri,
            'key': self.key
        }).encode("utf-8")

    @staticmethod
    def deserialize(data):
        try:
            if not isinstance(data, str):
                data = data.decode("utf-8")
            data = json.loads(data)
            if 'contents' not in data:
                raise ValueError("Missing 'contents' field.")
            if 'uri' not in data:
                raise ValueError("Missing 'uri' field.")
            if 'type' not in data:
                raise ValueError("Missing 'type' field.")
            order = Order(contents=data['contents'],
                          uri=data['uri'],
                          ty_pe=data['type'])
            if data['certificate_uri']:
                order.certificate_uri = data['certificate_uri']
            if data['key']:
                order.key = data['key']
            return order
        except (TypeError, ValueError, AttributeError) as e:
            raise IOError("Invalid account structure: {}".format(e))



#!/usr/bin/env python
#
# Copyright 2019 Flavio Garcia
# Copyright 2016-2017 Veeti Paananen under MIT License
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def confirm(msg, default=True):
    while True:
        choices = 'Y/n' if default else 'y/N'
        answer = input("{} [{}] ".format(msg, choices)).strip().lower()

        if answer in { 'yes', 'y' } or (default and not answer):
            return True
        elif answer in { 'no', 'n' } or (not default and not answer):
            return False



#!/usr/bin/env python
#
# Copyright 2019-2020 Flavio Garcia
# Copyright 2016-2017 Veeti Paananen under MIT License
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The certificate revocation command.
"""

import logging

from . import get_version
from .acme import AcmeV2
from .errors import AutomatoesError
from .crypto import (
    load_pem_certificate,
    get_certificate_domains
)
from .helpers import confirm

logger = logging.getLogger(__name__)


def revoke(server, account, certificate):
    print("Candango Automatoes {}. Manuale replacement.\n\n".format(
        get_version()))

    # Load the certificate
    try:
        with open(certificate, 'rb') as f:
            certificate = load_pem_certificate(f.read())
    except IOError as e:
        print("ERROR: Couldn't read the certificate.")
        raise AutomatoesError(e)

    # Confirm
    print("Are you sure you want to revoke this certificate? It includes the "
          "following domains:")
    for domain in get_certificate_domains(certificate):
        print("  {}".format(domain))
    if not confirm("This can't be undone. Confirm?", default=False):
        raise AutomatoesError("Aborting.")

    # Revoke.
    acme = AcmeV2(server, account)
    try:
        acme.revoke_certificate(certificate)
    except IOError as e:
        raise AutomatoesError(e)

    print("Certificate revoked.")



#!/usr/bin/env python
#
# Copyright 2019-2020 Flavio Goncalves Garcia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import get_version
from .crypto import (export_certificate_for_acme, generate_header, jose_b64,
                     sign_request, sign_request_v2)
from peasant.client import Peasant, PeasantTransport
import requests


class AcmeTransport(PeasantTransport):

    def __init__(self):
        self._default_header = {
            'User-Agent': ("automatoes %s (https://candango.org/p/automatoes)"
                           % get_version())
        }

    def set_directory(self):
        raise NotImplementedError

    def new_nonce(self):
        raise NotImplementedError

    def get(self, path, **kwargs):
        headers = kwargs.get("headers")
        verify = kwargs.get("verify")
        _headers = self._default_header.copy()
        if headers:
            _headers.update(headers)
        kwargs = {
            'headers': _headers
        }

        if verify:
            kwargs['verify'] = self.verify

        return requests.get(self.path(path), **kwargs)

    def head(self, path, **kwargs):
        raise NotImplementedError

    def post(self, path, body, **kwargs):
        headers = kwargs.get("headers")
        kid = kwargs.get("kid")
        verify = kwargs.get("verify")
        _headers = self._default_header.copy()
        _headers['Content-Type'] = "application/jose+json"
        if headers:
            _headers.update(headers)

        protected = self.get_headers(url=self.path(path))
        if kid:
            protected['kid'] = kid
            protected.pop('jwk')
        body = sign_request_v2(self.account.key, protected, body)
        kwargs = {
            'headers': _headers
        }
        if self.verify:
            kwargs['verify'] = self.verify
        return requests.post(self.path(path), data=body, **kwargs)

    def post_as_get(self, path, **kwargs):
        headers = kwargs.get("headers")
        kid = kwargs.get("kid")
        _headers = DEFAULT_HEADERS.copy()
        _headers['Content-Type'] = "application/jose+json"
        if headers:
            _headers.update(headers)

        protected = self.get_headers(url=self.path(path))
        protected['kid'] = kid
        protected.pop('jwk')
        body = sign_request_v2(self.account.key, protected, None)
        kwargs = {
            'headers': _headers
        }
        if self.verify:
            kwargs['verify'] = self.verify
        return requests.post(self.path(path), data=body, **kwargs)


class AcmePeasant(Peasant):

    def __init__(self, transport):
        super(AcmePeasant, self).__init__(transport)
        self._prior_version = 1
        self._version = 2




#!/usr/bin/env python
#
# Copyright 2019-2020 Flavio Garcia
# Copyright 2016-2017 Veeti Paananen under MIT License
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Cryptography, hopefully mostly correct.
"""

import logging

from cryptography import x509
from cryptography.x509 import NameOID
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.hazmat.primitives import hashes

import re

logger = logging.getLogger(__name__)


def create_csr(key, domains, must_staple=False):
    """
    Create a CSR in DER format for the specified key and domain names.
    """
    assert domains
    name = x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, domains[0]),
    ])
    san = x509.SubjectAlternativeName(
        [x509.DNSName(domain) for domain in domains])
    csr = (x509.CertificateSigningRequestBuilder()
           .subject_name(name).add_extension(san, critical=False))
    if must_staple:
        ocsp_must_staple = x509.TLSFeature(
            features=[x509.TLSFeatureType.status_request])
        csr = csr.add_extension(ocsp_must_staple, critical=False)
    return csr.sign(key, hashes.SHA256(), default_backend())


def load_csr(data):
    """
    Loads a PEM X.509 CSR.
    """
    return x509.load_pem_x509_csr(data, default_backend())


def load_der_certificate(data):
    """
    Loads a DER X.509 certificate.
    """
    return x509.load_der_x509_certificate(data, default_backend())


def load_pem_certificate(data):
    """
    Loads a PEM X.509 certificate.
    """
    return x509.load_pem_x509_certificate(data, default_backend())


def get_certificate_domain_name(cert):
    return cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value


def get_certificate_domains(cert):
    """
    Gets a list of all Subject Alternative Names in the specified certificate.
    """
    for ext in cert.extensions:
        ext = ext.value
        if isinstance(ext, x509.SubjectAlternativeName):
            return ext.get_values_for_type(x509.DNSName)
    return []


def export_pem_certificate(cert):
    """
    Exports a X.509 certificate as PEM.
    """
    return cert.public_bytes(Encoding.PEM)


def strip_certificates(data):
    p = re.compile("-----BEGIN CERTIFICATE-----\n(?s).+?"
                   "-----END CERTIFICATE-----\n")
    stripped_data = []
    for cert in p.findall(data.decode()):
        stripped_data.append(cert.encode())
    return stripped_data



#!/usr/bin/env python
#
# Copyright 2019-2020 Flavio Garcia
# Copyright 2016-2017 Veeti Paananen under MIT License
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The account upgrade command.
"""

from . import get_version
from .acme import AcmeV2
from .errors import AutomatoesError
from .helpers import confirm
from cartola import fs
import logging

logger = logging.getLogger(__name__)


def upgrade(server, account, account_path):
    acme_v2 = AcmeV2(server, account, upgrade=True)
    print("Candango Automatoes {}. Manuale replacement."
          "\n\n".format(get_version()))
    if acme_v2.is_uri_letsencrypt_acme_v1():
        print("Account's uri format is Let's Encrypt ACME V1.")
        print("Current uri: %s" % acme_v2.account.uri)
        if not confirm("Let's Encrypt ACME V2: %s.\nDo you want to "
                       "upgrade?" % acme_v2.letsencrypt_acme_uri_v1_to_v2()):
            raise AutomatoesError("Aborting.")
        account.uri = acme_v2.letsencrypt_acme_uri_v1_to_v2()
        fs.write(account_path, account.serialize(), binary=True)
        print("Account's uri upgraded to Let's Encrypt ACME V2.\n")
    else:
        print("Account's uri format isn't Let's Encrypt ACME V1.")
        print("Skipping upgrade action.")



#!/usr/bin/env python
#
# Copyright 2019-2020 Flavio Garcia
# Copyright 2016-2017 Veeti Paananen under MIT License
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Account registration.
"""

from . import get_version
from .acme import AcmeV2
from .errors import AutomatoesError, AccountAlreadyExistsError
from .crypto import (
    generate_rsa_key,
    load_private_key,
)
from .helpers import confirm
from .model import Account

import os


def register(server, account_path, email, key_file):
    print("Candango Automatoes {}. Manuale replacement.\n\n".format(
        get_version()))
    # Don't overwrite silently
    if os.path.exists(account_path):
        if not confirm("The account file {} already exists. Continuing will"
                       " overwrite it with the new key."
                       " Continue?".format(account_path), default=False):
            raise AutomatoesError("Aborting.")

    # Confirm e-mail
    if not confirm("You're about to register a new account with e-mail "
                   "{} as contact. Continue?".format(email)):
        raise AutomatoesError("Aborting.")

    # Load key or generate
    if key_file:
        try:
            with open(key_file, 'rb') as f:
                account = Account(key=load_private_key(f.read()))
        except (ValueError, AttributeError, TypeError, IOError) as e:
            print("ERROR: Couldn't read key.")
            raise AutomatoesError(e)
    else:
        print("Generating a new account key. This might take a second.")
        account = Account(key=generate_rsa_key(4096))
        print("  Key generated.")

    # Register
    acmev2 = AcmeV2(server, account)
    print("Registering...")
    try:
        terms_agreed = False
        print("  Retrieving terms of agreement ...")
        terms = acmev2.terms_from_directory()
        if terms is None:
            print("  There is no terms being enforced in this server. "
                  "Resuming execution...")
            terms_agreed = True
        else:
            print("  This server requires you to agree to these terms:")
            print("    {}".format(terms))
            if confirm("Agreed?"):
                terms_agreed = True
            else:
                print("Your account will still be created, but it won't be "
                      "usable before agreeing to terms.")
        acmev2.register(email, terms_agreed)
        print("  Account {} created.".format(account.uri))
    except IOError as e:
        print("ERROR: Registration failed due to a connection or request "
              "error.")
        raise AutomatoesError(e)

    # Write account
    directory = os.path.dirname(os.path.abspath(account_path))
    os.makedirs(directory, exist_ok=True)
    with open(account_path, 'wb') as f:
        os.chmod(account_path, 0o600)
        f.write(account.serialize())

    print("  Wrote account to {}.\n".format(account_path))
    print("What's next? Verify your domains with 'authorize' and use 'issue' "
          "to get new certificates.")



#!/usr/bin/env python
#
# Copyright 2019-2020 Flavio Garcia
# Copyright 2016-2017 Veeti Paananen under MIT License
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
ACME API client.
"""

from . import get_version
from .crypto import (export_certificate_for_acme, generate_header, jose_b64,
                     sign_request, sign_request_v2)
from .errors import AccountAlreadyExistsError, AcmeError
from .model import Order, Challenge
from collections import namedtuple
import copy
import datetime
import hashlib
import logging
import requests
import time
from urllib.parse import urljoin, urlparse

logger = logging.getLogger(__name__)

DEFAULT_HEADERS = {
    'User-Agent': "automatoes {} (https://candango.org/p/automatoes)".format(
        get_version()),
}


class Acme:

    def __init__(self, url, account, directory="directory", verify=None):
        self.url = url
        self.account = None
        self.directory = directory
        self.directory_cache = None
        self.verify = verify
        self.set_account(account)

    def set_account(self, account):
        self.account = account

    @property
    def key(self):
        if self.account is None:
            return None
        return self.account.key

    def get_directory(self):
        return self.get("/directory")

    def get_nonce(self):
        """
        Gets a new nonce.
        """
        return self.get('/directory').headers.get('Replay-Nonce')

    def get_headers(self):
        """
        Builds a new pair of headers for signed requests.
        """
        header = generate_header(self.account.key)
        protected_header = copy.deepcopy(header)
        protected_header['nonce'] = self.get_nonce()
        return header, protected_header

    def register(self, email):
        """
        Registers the current account on the server.
        """
        response = self.post("/acme/new-reg", {
            'resource': "new-reg",
            'contact': [
                "mailto:{}".format(email)
            ],
        })
        uri = response.headers.get("Location")
        if response.status_code == 201:
            self.account.uri = uri

            # Find terms of service from link headers
            terms = response.links.get("terms-of-service")

            return RegistrationResult(
                contents=_json(response),
                uri=uri,
                terms=(terms['url'] if terms else None)
            )
        elif response.status_code == 409:
            raise AccountAlreadyExistsError(response, uri)
        raise AcmeError(response)

    def get_registration(self):
        """
        Gets available account information from the server.
        """
        response = self.post(self.account.uri, {
            'resource': "reg",
        })
        if str(response.status_code).startswith("2"):
            return _json(response)
        raise AcmeError(response)

    def update_registration(self, params=None):
        """
        Updates registration information on the server.
        """
        params = params or {}
        params['resource'] = "reg"

        response = self.post(self.account.uri, params)
        if str(response.status_code).startswith("2"):
            return True
        raise AcmeError(response)

    def new_authorization(self, domain, type='dns'):
        """
        Requests a new authorization for the specified domain.
        :return Order
        """
        response = self.post('/acme/new-authz', {
            'resource': "new-authz",
            'identifier': {'type': type, 'value': domain}
        })
        if response.status_code == 201:
            return NewAuthorizationResult(_json(response),
                                          response.headers.get('Location'))
        raise AcmeError(response)

    def validate_authorization(self, uri, _type, key_authorization=None):
        """
        Marks the specified validation as complete.
        """
        response = self.post(uri, {
            'resource': "challenge",
            'type': _type,
            'keyAuthorization': key_authorization,
        })
        if str(response.status_code).startswith('2'):
            return True
        raise AcmeError(response)

    def get_authorization(self, uri):
        """
        Returns the authorization status.
        """
        response = self.get(uri)
        try:
            return response.json()
        except (ValueError, TypeError, AttributeError) as e:
            raise AcmeError(e)

    def issue_certificate(self, csr):
        http_headers = {'Accept': "application/pkix-cert"}
        response = self.post('/acme/new-cert', {
            'resource': "new-cert",
            'csr': csr,
        }, headers=http_headers)
        if response.status_code == 201:
            # Get the issuer certificate
            chain = response.links.get("up")
            if chain:
                chain = requests.get(chain['url'],
                                     headers=DEFAULT_HEADERS).content
            return IssuanceResult(
                response.content,
                response.headers.get("Location"),
                chain,
            )
        raise AcmeError(response)

    def revoke_certificate(self, cert):
        response = self.post('/acme/revoke-cert', {
            'resource': "revoke-cert",
            'certificate': cert,
        })
        if response.status_code == 200:
            return True
        raise AcmeError(response)

    def get(self, path, headers=None):
        _headers = DEFAULT_HEADERS.copy()
        if headers:
            _headers.update(headers)
        kwargs = {
            'headers': _headers
        }

        if self.verify:
            kwargs['verify'] = self.verify

        return requests.get(self.path(path), **kwargs)

    def post(self, path, body, headers=None):
        _headers = DEFAULT_HEADERS.copy()
        _headers['Content-Type'] = "application/json"
        if headers:
            _headers.update(headers)

        header, protected = self.get_headers()
        body = sign_request(self.account.key, header, protected, body)

        kwargs = {
            'headers': _headers
        }

        if self.verify:
            kwargs['verify'] = self.verify

        return requests.post(self.path(path), data=body, **kwargs)

    def path(self, path):
        # If https is in we assume that was returned by the directory
        if path.startswith("https"):
            return path
        # Make sure path is relative
        if path.startswith("http"):
            path = urlparse(path).path
        url_parsed = urlparse(self.url)
        if url_parsed.path != "":
            url_parsed_path = url_parsed.path
            if url_parsed_path.startswith("/"):
                url_parsed_path = url_parsed_path[1:]
            if path.startswith("/"):
                path = path[1:]
            path = "%s/%s" % (url_parsed_path, path)
        return urljoin(self.url, path)


RegistrationResult = namedtuple("RegistrationResult", "contents uri terms")
NewAuthorizationResult = namedtuple("NewAuthorizationResult", "contents uri")
IssuanceResult = namedtuple("IssuanceResult",
                            "certificate location intermediate")


class AcmeV2(Acme):

    def __init__(self, url, account, directory="directory", verify=None,
                 upgrade=False):
        super(AcmeV2, self).__init__(url, account, directory, verify)
        if self.is_uri_letsencrypt_acme_v1():
            if not upgrade:
                logger.warning("WARNING: The account is using Let's Encrypt "
                               "discontinued ACME V1 url.")
                logger.warning("WARNING: Upgrading account to ACME V2 "
                               "temporally.")
                logger.warning("WARNING: Please run 'manuale upgrade' to make "
                               "the change permanently.")
                self.account.uri = self.letsencrypt_acme_uri_v1_to_v2()

    def is_uri_letsencrypt_acme_v1(self):
        if self.account is None:
            return False
        if self.account.uri is None:
            return False
        return "acme-v01.api.letsencrypt.org" in self.account.uri

    def letsencrypt_acme_uri_v1_to_v2(self):
        if self.account is None:
            return None
        uri = self.account.uri.replace(
            "acme-v01.api.letsencrypt.org",
            "acme-v02.api.letsencrypt.org"
        )
        return uri.replace("acme/reg", "acme/acct")

    def head(self, path, headers=None):
        _headers = DEFAULT_HEADERS.copy()
        if headers:
            _headers.update(headers)

        kwargs = {
            'headers': _headers
        }
        if self.verify:
            kwargs['verify'] = self.verify
        return requests.head(self.path(path), **kwargs)

    def get_headers(self, url=None):
        """
        Builds a new pair of headers for signed requests.
        """
        header = generate_header(self.account.key)
        protected_header = copy.deepcopy(header)
        protected_header['nonce'] = self.get_nonce()
        if url is not None:
            protected_header['url'] = url
        return protected_header

    def get_directory(self):
        if not self.directory_cache:
            self.directory_cache = self.get("/{}".format(self.directory))
        return self.directory_cache

    def url_from_directory(self, what_url):
        response = self.get_directory()
        if response.status_code == 200:
            return response.json()[what_url]
        return None

    def terms_from_directory(self):
        response = self.get_directory()
        if response.status_code == 200:
            if "meta" in response.json():
                if "termsOfService" in response.json()['meta']:
                    return response.json()['meta']['termsOfService']
        return None

    def get_nonce(self):
        """ Gets a new nonce.
        """
        return self.head(self.url_from_directory('newNonce'), {
            'resource': "new-reg",
            'payload': None,
        }).headers.get('Replay-Nonce')

    def register(self, email, terms_agreed=False):
        """Registers the current account on the server.
        """
        payload = {
           "termsOfServiceAgreed": terms_agreed,
           "contact": [
             "mailto:{email}".format(email=email)
           ]
         }
        response = self.post(
            self.url_from_directory('newAccount'),
            payload
        )

        uri = response.headers.get("Location")

        if response.status_code == 201:
            self.account.uri = uri

            # Find terms of service from link headers
            terms = self.terms_from_directory()

            return RegistrationResult(
                contents=_json(response),
                uri=uri,
                terms=terms
            )
        elif response.status_code == 409:
            raise AccountAlreadyExistsError(response, uri)
        raise AcmeError(response)

    def get_registration(self):
        """
        Get available account information from the server.
        """
        response = self.post_as_get(self.account.uri, kid=self.account.uri)
        if str(response.status_code).startswith("2"):
            return _json(response)
        raise AcmeError(response)

    def new_order(self, domains, type='dns'):
        """ Requests a new authorization for the specified domain.
        """
        domains_with_type = []
        if not isinstance(domains, list):
            domains = [domains]
        for domain in domains:
            domains_with_type.append({'type': 'dns', 'value': domain})
        response = self.post(self.url_from_directory('newOrder'), {
            'identifiers': domains_with_type
        }, kid=self.account.uri)
        if response.status_code == 201:
            return Order(
                contents=_json(response),
                uri=response.headers.get('Location'),
                ty_pe=type
            )
        raise AcmeError(response)

    def query_orders(self):
        """ Query existent order status


        :param Order order: order to be challenged
        :return: Order
        """
        raise NotImplementedError

    def query_order(self, order):
        """ Query existent order status
        :param Order order: order to be challenged
        :return: Order
        """
        response = self.post_as_get(order.uri, kid=self.account.uri)
        if response.status_code == 200:
            return Order(
                contents=_json(response),
                uri=order.uri,
                ty_pe=order.type
            )
        raise AcmeError(response)

    def get_order_challenges(self, order):
        """ Return all challenges from an order .
        :param Order order: order to be challenged
        :return: Order
        """
        domains = [identifier['value'] for identifier in
                   order.contents['identifiers']]
        order_challenges = []
        for auth in order.contents['authorizations']:
            auth_response = _json(self.post_as_get(auth, self.account.uri))
            for challenge in auth_response['challenges']:
                key_authorization = "{}.{}".format(
                    challenge['token'], self.account.thumbprint)
                digest = hashlib.sha256()
                digest.update(key_authorization.encode('ascii'))
                if order.type in challenge['type']:
                    if "expires" not in auth_response:
                        expire_date = (datetime.datetime.now() +
                                       datetime.timedelta(days=2))
                        auth_response['expires'] = expire_date.strftime(
                            "%Y-%m-%dT%H:%M:%S")
                    order_challenges.append(Challenge(
                        contents=challenge,
                        domain=auth_response['identifier']['value'],
                        expires=auth_response['expires'],
                        status=auth_response['status'],
                        ty_pe=order.type,
                        key=jose_b64(digest.digest())
                    ))
        return order_challenges

    def verify_order_challenge(self, challenge, timeout=5, retry_limit=5):
        """ Return all challenges from an order .
        :param Challenge challenge: A challenge from the order
        :param int timeout: timeout before check challenge status
        :param retry_limit: retry limit of checks of a challenge
        :return:
        """
        parsed_url = urlparse(self.url)
        host = parsed_url.hostname
        if parsed_url.port:
            host = "{}:{}".format(host, parsed_url.port)

        response = _json(self.post(challenge.contents['url'],
                                   {},
                                   {'Host': host},
                                   kid=self.account.uri))
        retries = 0
        while response['status'] == "pending":
            if retries < retry_limit:
                time.sleep(timeout)
                response = _json(self.post_as_get(challenge.contents['url'],
                                                  kid=self.account.uri))
        return response

    def finalize_order(self, order, csr):
        """
        Marks the specified validation as complete.
        :param OrderResult order: authorization to be
        validated
        :return:
        """
        response = self.post(order.contents['finalize'], {
            'csr': export_certificate_for_acme(csr),
        }, kid=self.account.uri)
        if _json(response)['status'] == "valid":
            order.certificate_uri = _json(response)['certificate']
        if response.status_code == 200:
            return _json(response)
        raise AcmeError(response)

    def await_for_order_fulfillment(self, order, timeout=2, iterations=5):
        response = self.post_as_get(order.uri, kid=self.account.uri)
        iteration_count = 0
        while _json(response)['status'] != "valid":
            if iteration_count == iterations:
                break
            time.sleep(timeout)
            response = self.post_as_get(order.uri,
                                        kid=self.account.uri)
            iteration_count += 1

        if _json(response)['status'] == "valid":
            order.certificate_uri = _json(response)['certificate']

        if response.status_code == 200:
            return _json(response)
        raise AcmeError(response)

    def download_order_certificate(self, order):
        response = self.post_as_get(order.certificate_uri,
                                    kid=self.account.uri)
        if response.status_code == 200:
            order.certificate = response.content
            return response
        raise AcmeError(response)

    def revoke_certificate(self, cert):
        response = self.post(self.url_from_directory('revokeCert'), {
            'certificate':  export_certificate_for_acme(cert),
        }, kid=self.account.uri)
        if response.status_code == 200:
            return response
        raise AcmeError(response)

    def post(self, path, body, headers=None, kid=None):
        _headers = DEFAULT_HEADERS.copy()
        _headers['Content-Type'] = "application/jose+json"
        if headers:
            _headers.update(headers)

        protected = self.get_headers(url=self.path(path))
        if kid:
            protected['kid'] = kid
            protected.pop('jwk')
        body = sign_request_v2(self.account.key, protected, body)
        kwargs = {
            'headers': _headers
        }
        if self.verify:
            kwargs['verify'] = self.verify
        return requests.post(self.path(path), data=body, **kwargs)

    def post_as_get(self, path, kid, headers=None):
        _headers = DEFAULT_HEADERS.copy()
        _headers['Content-Type'] = "application/jose+json"
        if headers:
            _headers.update(headers)

        protected = self.get_headers(url=self.path(path))
        protected['kid'] = kid
        protected.pop('jwk')
        body = sign_request_v2(self.account.key, protected, None)
        kwargs = {
            'headers': _headers
        }
        if self.verify:
            kwargs['verify'] = self.verify
        return requests.post(self.path(path), data=body, **kwargs)


def _json(response):
    try:
        return response.json()
    except ValueError as e:
        raise AcmeError("Invalid JSON response. {}".format(e))



#!/usr/bin/env python
#
# Copyright 2019-2020 Flavio Garcia
# Copyright 2016-2017 Veeti Paananen under MIT License
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The domain authorization command.
"""

from . import get_version
from .acme import AcmeV2
from .crypto import generate_jwk_thumbprint
from .errors import AutomatoesError
from .model import Order

from cartola import fs, sysexits
import hashlib
import os
import sys


def create_order(acme, domains, method, order_file):
    order = acme.new_order(domains, method)
    update_order(order, order_file)
    return order


def update_order(order, order_file):
    fs.write(order_file, order.serialize().decode())


def clean_http_challenges(files):
    # Clean up created files
    for path in files:
        try:
            os.remove(path)
        except:
            print("Couldn't delete http challenge file {}".format(path))


def clean_challenge_file(challenge_file):
    try:
        os.remove(challenge_file)
    except:
        print("Couldn't delete challenge file {}".format(challenge_file))


def authorize(server, paths, account, domains, method, verbose=False):
    print("Candango Automatoes {}. Manuale replacement.\n\n".format(
        get_version()))

    current_path = paths['current']
    orders_path = paths['orders']
    domains_hash = hashlib.sha256(
        "_".join(domains).encode('ascii')).hexdigest()
    order_path = os.path.join(orders_path, domains_hash)
    order_file = os.path.join(order_path, "order.json".format(domains_hash))

    if not os.path.exists(orders_path):
        if verbose:
            print("Orders path not found creating it at {}."
                  "".format(orders_path))
        os.mkdir(orders_path)
        os.chmod(orders_path, 0o770)
    else:
        if verbose:
            print("Orders path found at {}.".format(orders_path))

    if not os.path.exists(order_path):
        if verbose:
            print("Current order {} path not found creating it at orders "
                  "path.\n".format(domains_hash))
        os.mkdir(order_path)
        os.chmod(order_path, 0o770)
    else:
        if verbose:
            print("Current order {} path found at orders path.\n".format(
                domains_hash))

    method = method
    acme = AcmeV2(server, account)

    try:
        print("Authorizing {}.\n".format(", ".join(domains)))
        # Creating orders for domains if not existent
        if not os.path.exists(order_file):
            if verbose:
                print("  Order file not found creating it.")
            order = create_order(acme, domains, method, order_file)
        else:
            if verbose:
                print("  Found order file. Querying ACME server for current "
                      "status.")
            order = Order.deserialize(fs.read(order_file))
            try:
                server_order = acme.query_order(order)
                order.contents = server_order.contents
            except:
                print("    WARNING: Old order. Setting it as expired.\n")
                order.contents['status'] = "expired"
            update_order(order, order_file)

            if not order.expired and not order.invalid:
                if order.contents['status'] == 'valid':
                    print("  Order is valid and expires at {}. Please run "
                          "the issue "
                          "command.\n".format(order.contents['expires']))
                    print("  {} domain(s) authorized. Let's Encrypt!".format(
                        len(domains)))
                    sys.exit(sysexits.EX_OK)
                else:
                    if verbose:
                        print("    Order still pending and expires "
                              "at {}.\n".format(order.contents['expires']))
            else:
                if order.invalid:
                    print("    WARNING: Invalid order, renewing it.\n    Just "
                          "continue with the authorization when all "
                          "verifications are in place.\n")
                else:
                    print("  WARNING: Expired order. Renewing order.\n")
                os.remove(order_file)
                order = create_order(acme, domains, method, order_file)
                update_order(order, order_file)

        pending_challenges = []

        for challenge in acme.get_order_challenges(order):
            print("  Requesting challenge for {}.".format(challenge.domain))
            if challenge.status == 'valid':
                print("    {} is already authorized until {}.".format(
                    challenge.domain, challenge.expires))
                continue
            else:
                challenge_file = os.path.join(order_path, challenge.file_name)
                if verbose:
                    print("    Creating challenge file {}.\n".format(
                        challenge.file_name))
                fs.write(challenge_file, challenge.serialize().decode())
                pending_challenges.append(challenge)

        # Quit if nothing to authorize
        if not pending_challenges:
            print("\nAll domains are already authorized, exiting.")
            sys.exit(sysexits.EX_OK)

        files = set()
        if method == 'dns':
            print("\n  DNS verification required. Make sure these TXT records"
                  " are in place:\n")
            for challenge in pending_challenges:
                print("    _acme-challenge.{}.  IN TXT  "
                      "\"{}\"".format(challenge.domain, challenge.key))
        elif method == 'http':
            print("\n  HTTP verification required. Make sure these files are "
                  "in place:\n")
            for challenge in pending_challenges:
                token = challenge.contents['token']
                # path sanity check
                assert (token and os.path.sep not in token and '.' not in
                        token)
                files.add(token)
                fs.write(
                    os.path.join(current_path, token),
                    "%s.%s" % (token,
                               generate_jwk_thumbprint(account.key))
                )
                print("    http://{}/.well-known/acme-challenge/{}".format(
                    challenge.domain, token))

            print("\n  The necessary files have been written to the current "
                  "directory.\n")
        # Wait for the user to complete the challenges
        input("\nPress Enter to continue.\n")

        # Validate challenges
        done, failed, pending = set(), set(), set()
        for challenge in pending_challenges:
            print("  {}: waiting for verification. Checking in 5 "
                  "seconds.".format(challenge.domain))
            response = acme.verify_order_challenge(challenge, 5, 1)
            if response['status'] == "valid":
                print("  {}: OK! Authorization lasts until {}.".format(
                    challenge.domain, challenge.expires))
                done.add(challenge.domain)
            elif response['status'] == 'invalid':
                print("  {}: {} ({})".format(
                    challenge.domain,
                    response['error']['detail'],
                    response['error']['type'])
                )
                failed.add(challenge.domain)
                break
            else:
                print("{}: Pending!".format(challenge.domain))
                pending.add(challenge.domain)
                break

        challenge_file = os.path.join(order_path, challenge.file_name)
        # Print results
        if failed:
            print("  {} domain(s) authorized, {} failed.".format(
                    len(done),
                    len(failed),
            ))
            print("  Authorized: {}".format(' '.join(done) or "N/A"))
            print("  Failed: {}".format(' '.join(failed)))
            print("  WARNING: The current order will be invalidated. "
                  "Try again.")
            if verbose:
                print("    Deleting invalid challenge file {}.\n".format(
                    challenge.file_name))
            clean_challenge_file(challenge_file)
            os.remove(order_file)
            os.rmdir(order_path)
            if method == 'http':
                print(files)
                clean_http_challenges(files)
            sys.exit(sysexits.EX_FATAL_ERROR)
        else:
            if pending:
                print("  {} domain(s) authorized, {} pending.".format(
                    len(done),
                    len(pending)))
                print("  Authorized: {}".format(' '.join(done) or "N/A"))
                print("  Pending: {}".format(' '.join(pending)))
                print("  Try again.")
                sys.exit(sysexits.EX_CANNOT_EXECUTE)
            else:
                if verbose:
                    print("    Deleting valid challenge file {}.".format(
                        challenge.file_name))
                clean_challenge_file(challenge_file)
                if verbose:
                    print("    Querying ACME server for current status.\n")
                server_order = acme.query_order(order)
                order.contents = server_order.contents
                update_order(order, order_file)
                print("  {} domain(s) authorized. Let's Encrypt!".format(
                    len(done)))
        if method == 'http':
            clean_http_challenges(files)
        sys.exit(sysexits.EX_OK)
    except IOError as e:
        print("A connection or service error occurred. Aborting.")
        raise AutomatoesError(e)



#!/usr/bin/env python
#
# Copyright 2019-2020 Flavio Garcia
# Copyright 2016-2017 Veeti Paananen under MIT License
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The command line interface.
"""
from . import get_version
from .authorize import authorize
from .issue import issue
from .info import info
from .model import Account
from .register import register
from .revoke import revoke
from .upgrade import upgrade
from .errors import AutomatoesError

import argparse
from cartola import sysexits
import logging
import sys
import os

logger = logging.getLogger(__name__)

# Text
DESCRIPTION = \
"""
Candango Automatoes {}. Manuale replacement.

Interact with ACME certification authorities such as Let's Encrypt.

No idea what you're doing? Register an account, authorize your domains and
issue a certificate or two. Call a command with -h for more instructions.
""".format(get_version())

DESCRIPTION_REGISTER = \
"""
Creates a new account key and registers on the server. The resulting --account
is saved in the specified file, and required for most other operations.

You only have to do this once. Keep the account file safe and secure: it
contains your private key, and you need it to get certificates!
"""

DESCRIPTION_AUTHORIZE = \
"""
Authorizes a domain or multiple domains for your account through DNS or HTTP
verification. You will need to set up DNS records or HTTP files as prompted.

After authorizing a domain, you can issue certificates for it. Authorizations
can last for a long time, so you might not need to do this every time you want
a new certificate.  This depends on the server being used. You should see an
expiration date for the authorization after completion.

If a domain is already authorized, the authorization's expiration date will be
printed.
"""

DESCRIPTION_ISSUE = \
"""
Issues a certificate for one or more domains. Hopefully needless to say, you
must have valid authorizations for the domains you specify first.

This will generate a new RSA key and CSR for you. But if you want, you can
bring your own with the --key-file and --csr-file attributes. You can also set
a custom --key-size. (Don't try something stupid like 512, the server won't
accept it. I tried.)

The resulting key and certificate are written into domain.pem and domain.crt.
A chained certificate with the intermediate included is also written to
domain.chain.crt. You can change the --output directory to something else from
the working directory as well.

(If you're passing your own CSR, the given domains can be whatever you want.)

Note that unlike many other certification authorities, ACME does not add a
non-www or www alias to certificates. If you want this to happen, add it
yourself. You need to authorize both as well.

Certificate issuance has a server-side rate limit. Don't overdo it.
"""

DESCRIPTION_REVOKE = \
"""
Revokes a certificate. The certificate must have been issued using the
current account.
"""

DESCRIPTION_INFO = \
"""
Display registration info for the current account.
"""

DESCRIPTION_UPGRADE = \
"""
Upgrade current account's uri from Let's Encrypt ACME V1 to ACME V2.
"""

# Defaults
LETS_ENCRYPT_PRODUCTION = "https://acme-v02.api.letsencrypt.org/"
DEFAULT_ACCOUNT_PATH = 'account.json'
DEFAULT_CERT_KEY_SIZE = 4096


# Command handlers
def _register(args):
    register(
        server=args.server,
        account_path=args.account,
        email=args.email,
        key_file=args.key_file
    )


def _authorize(args):
    paths = get_paths(args.account)
    account = load_account(args.account)
    verbose = False
    if args.verbose > 0:
        verbose = True
    authorize(args.server, paths, account, args.domain, args.method, verbose)


def _issue(args):
    paths = get_paths(args.account)
    account = load_account(args.account)
    verbose = False
    if args.verbose > 0:
        verbose = True
    issue(
        server=args.server,
        paths=paths,
        account=account,
        domains=args.domain,
        key_size=args.key_size,
        key_file=args.key_file,
        csr_file=args.csr_file,
        output_path=args.output,
        must_staple=args.ocsp_must_staple,
        verbose=verbose
    )


def _revoke(args):
    account = load_account(args.account)
    revoke(
        server=args.server,
        account=account,
        certificate=args.certificate
    )


def _info(args):
    paths = get_paths(args.account)
    account = load_account(args.account)
    info(args.server, account, paths)


def _upgrade(args):
    account_path = args.account
    account = load_account(args.account)
    upgrade(args.server, account, account_path)


def get_paths(account_file):
    current_path = os.path.dirname(os.path.abspath(account_file))
    return {
        'authorizations': os.path.join(current_path, "authorizations"),
        'current': current_path,
        'orders': os.path.join(current_path, "orders"),
    }


def get_meta_paths(path):
    return {
        'orders': os.path.join(path, "orders"),
        'authorizations': os.path.join(path, "authorizations"),
    }


def load_account(path):
    # Show a more descriptive message if the file doesn't exist.
    if not os.path.exists(path):
        logger.error("Couldn't find an account file at {}.".format(path))
        logger.error("Are you in the right directory? Did you register yet?")
        logger.error("Run 'automatoes -h' for instructions.")
        raise AutomatoesError()

    try:
        with open(path, 'rb') as f:
            return Account.deserialize(f.read())
    except (ValueError, IOError) as e:
        logger.error("Couldn't read account file. Aborting.")
        raise AutomatoesError(e)


class Formatter(argparse.ArgumentDefaultsHelpFormatter,
                argparse.RawDescriptionHelpFormatter):
    pass


def automatoes_main():
    print("The automatoes command is not implemented yet.")


# Where it all begins.
def manuale_main():
    parser = argparse.ArgumentParser(
        description=DESCRIPTION,
        formatter_class=Formatter,
    )
    subparsers = parser.add_subparsers()

    # Server switch
    parser.add_argument('--server', '-s', help="The ACME server to use",
                        default=LETS_ENCRYPT_PRODUCTION)
    parser.add_argument('--account', '-a',
                        help="The account file to use or create",
                        default=DEFAULT_ACCOUNT_PATH)

    # Verbosity
    parser.add_argument('--verbose', '-v', action="count",
                        help="Set verbose mode", default=0)

    # Account creation
    register = subparsers.add_parser(
        'register',
        help="Create a new account and register",
        description=DESCRIPTION_REGISTER,
        formatter_class=Formatter,
    )
    register.add_argument('email', type=str, help="Account e-mail address")
    register.add_argument('--key-file', '-k',
                          help="Existing key file to use for the account")
    register.set_defaults(func=_register)

    # Domain verification
    authorize = subparsers.add_parser(
        'authorize',
        help="Verify domain ownership",
        description=DESCRIPTION_AUTHORIZE,
        formatter_class=Formatter,
    )
    authorize.add_argument('domain',
                           help="One or more domain names to authorize",
                           nargs='+')
    authorize.add_argument('--method',
                           '-m',
                           help="Authorization method",
                           choices=('dns', 'http'),
                           default='dns')
    authorize.set_defaults(func=_authorize)

    # Certificate issuance
    issue = subparsers.add_parser(
        'issue',
        help="Request a new certificate",
        description=DESCRIPTION_ISSUE,
        formatter_class=Formatter,
    )
    issue.add_argument(
        'domain',
        help="One or more domain names to include in the certificate",
        nargs='+')
    issue.add_argument('--key-size', '-b',
                       help="The key size to use for the certificate",
                       type=int, default=DEFAULT_CERT_KEY_SIZE)
    issue.add_argument('--key-file', '-k',
                       help="Existing key file to use for the certificate")
    issue.add_argument('--csr-file', help="Existing signing request to use")
    issue.add_argument('--output', '-o',
                       help="The output directory for created objects",
                       default='.')
    issue.add_argument('--ocsp-must-staple',
                       dest='ocsp_must_staple',
                       help="CSR: Request OCSP Must-Staple extension",
                       action='store_true')
    issue.add_argument('--no-ocsp-must-staple',
                       dest='ocsp_must_staple',
                       help=argparse.SUPPRESS,
                       action='store_false')
    issue.set_defaults(func=_issue, ocsp_must_staple=False)

    # Certificate revocation
    revoke = subparsers.add_parser(
        'revoke',
        help="Revoke an issued certificate",
        description=DESCRIPTION_REVOKE,
        formatter_class=Formatter,
    )
    revoke.add_argument("certificate", help="The certificate file to revoke")
    revoke.set_defaults(func=_revoke)

    # Account info
    info = subparsers.add_parser(
        'info',
        help="Display account information",
        description=DESCRIPTION_INFO,
        formatter_class=Formatter,
    )
    info.set_defaults(func=_info)

    # Account upgrade
    upgrade = subparsers.add_parser(
        'upgrade',
        help="Upgrade account's uri from Let's Encrypt ACME V1 to V2",
        description=DESCRIPTION_UPGRADE,
        formatter_class=Formatter,
    )
    upgrade.set_defaults(func=_upgrade)

    # Version
    version = subparsers.add_parser("version", help="Show the version number")
    version.set_defaults(func=lambda *args: logger.info(
        "automatoes {}\n\nThis tool is a full manuale "
        "replacement.\nJust run manuale instead of automatoes"
        ".".format(get_version())))

    # Parse
    args = parser.parse_args()
    if not hasattr(args, 'func'):
        parser.print_help()
        sys.exit(sysexits.EX_MISUSE)

    # Set up logging
    root = logging.getLogger('automatoes')
    root.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(logging.Formatter("%(message)s"))
    root.addHandler(handler)

    # Let's encrypt
    try:
        args.func(args)
    except AutomatoesError as e:
        if str(e):
            logger.error(e)
        sys.exit(sysexits.EX_SOFTWARE)
    except KeyboardInterrupt:
        logger.error("")
        logger.error("Interrupted.")
        sys.exit(sysexits.EX_TERMINATED_BY_CRTL_C)
    except Exception as e:
        logger.error("Oops! An unhandled error occurred. Please file a bug.")
        logger.exception(e)
        sys.exit(sysexits.EX_CATCHALL)



#!/usr/bin/env python
#
# Copyright 2019-2020 Flavio Garcia
# Copyright 2016-2017 Veeti Paananen under MIT License
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The moment you've been waiting for: actually getting SSL. For free!
"""

from . import get_version
from .acme import AcmeV2
from .authorize import update_order
from .crypto import (
    generate_rsa_key,
    load_private_key,
    export_private_key,
    create_csr,
    load_csr,
    export_csr_for_acme,
    load_pem_certificate,
    export_pem_certificate,
    strip_certificates,
)
from .errors import AutomatoesError
from .model import Order

import binascii
from cartola import fs, sysexits
from cryptography.hazmat.primitives.hashes import SHA256
import hashlib
import logging
import os
import sys

logger = logging.getLogger(__name__)

EXPIRATION_FORMAT = "%Y-%m-%d"


def issue(server, paths, account, domains, key_size, key_file=None,
          csr_file=None, output_path=None, must_staple=False, verbose=False):
    print("Candango Automatoes {}. Manuale replacement.\n\n".format(
        get_version()))

    current_path = paths['current']
    orders_path = paths['orders']
    domains_hash = hashlib.sha256(
        "_".join(domains).encode('ascii')).hexdigest()
    order_path = os.path.join(orders_path, domains_hash)
    order_file = os.path.join(order_path, "order.json".format(domains_hash))

    if not os.path.exists(orders_path):
        print(" ERROR: Orders path not found. Please run before: manuale "
              "authorize {}".format(" ".join(domains)))
        sys.exit(sysexits.EX_CANNOT_EXECUTE)
    else:
        if verbose:
            print("Orders path found at {}.".format(orders_path))

    if verbose:
        print("Searching order file {}.".format(order_file))

    if not os.path.exists(order_path):
        print(" ERROR: Order file not found. Please run before: manuale "
              "authorize {}".format(" ".join(domains)))
        sys.exit(sysexits.EX_CANNOT_EXECUTE)
    else:
        if verbose:
            print("Current order {} path found at orders path.\n".format(
                domains_hash))

    acme = AcmeV2(server, account)
    order = Order.deserialize(fs.read(order_file))
    if order.contents['status'] == "pending":
        if verbose:
            print("Querying ACME server for current status.")
        server_order = acme.query_order(order)
        order.contents = server_order.contents
        update_order(order, order_file)
        if order.contents['status'] in ["pending", "invalid"]:
            print(" ERROR: Order not ready or invalid. Please re-run: manuale"
                  " authorize {}.".format(" ".join(domains)))
            sys.exit(sysexits.EX_CANNOT_EXECUTE)
    elif order.contents['status'] == "invalid":
        print(" ERROR: Invalid order. Please re-run: manuale authorize "
              "{}.".format(" ".join(domains)))
        sys.exit(sysexits.EX_CANNOT_EXECUTE)

    if not output_path or output_path == '.':
        output_path = os.getcwd()

    # Load key if given
    if key_file:
        try:
            with open(key_file, 'rb') as f:
                certificate_key = load_private_key(f.read())
            order.key = export_private_key(certificate_key).decode('ascii')
            update_order(order, order_file)
        except (ValueError, AttributeError, TypeError, IOError) as e:
            print("ERROR: Couldn't read certificate key.")
            raise AutomatoesError(e)
    else:
        certificate_key = None

    # Load CSR or generate
    if csr_file:
        try:
            with open(csr_file, 'rb') as f:
                csr = export_csr_for_acme(load_csr(f.read()))
        except (ValueError, AttributeError, TypeError, IOError) as e:
            print("ERROR: Couldn't read CSR.")
            raise AutomatoesError(e)
    else:
        # Generate key
        if not key_file:
            if order.key is None:
                print("Generating a {} bit RSA key. This might take a "
                      "second.".format(key_size))
                certificate_key = generate_rsa_key(key_size)
                print("  Key generated.")
                order.key = export_private_key(certificate_key).decode('ascii')
                update_order(order, order_file)
                print("  Order updated with generated key.")
            else:
                print("Previous RSA key found in the order. Loading the key.")
                certificate_key = load_private_key(order.key.encode('ascii'))

        csr = create_csr(certificate_key, domains, must_staple=must_staple)

    try:
        logger.info("Requesting certificate issuance...")
        if order.contents['status'] == "ready":
            final_order = acme.finalize_order(order, csr)
            order.contents = final_order
            update_order(order, order_file)
            if final_order['status'] in ["processing", "valid"]:
                if verbose:
                    print("  Order {} finalized. Certificate is being "
                          "issued.".format(domains_hash))
            else:
                print(" ERROR: Order not ready or invalid. Please re-run: "
                      "manuale authorize {}.".format(" ".join(domains)))
                sys.exit(sysexits.EX_CANNOT_EXECUTE)
        elif order.contents['status'] in ["valid", "processing"]:
            print("  Order {} is already processing or valid. Downloading "
                  "certificate.".format(domains_hash))
        else:
            print(" ERROR: Order not ready or invalid. Please re-run: manuale "
                  "authorize {}.".format(" ".join(domains)))
            sys.exit(sysexits.EX_CANNOT_EXECUTE)

        if order.certificate_uri is None:
            if verbose:
                print("  Checking order {} status.".format(domains_hash))
            fulfillment = acme.await_for_order_fulfillment(order)
            if fulfillment['status'] == "valid":
                order.contents = fulfillment
                update_order(order, order_file)
            else:
                print(" ERROR: Order not ready or invalid. Please re-run: "
                      "manuale authorize {}.".format(" ".join(domains)))
                sys.exit(sysexits.EX_CANNOT_EXECUTE)
        else:
            print("  We already know the certificate uri for order {}. "
                  "Downloading certificate.".format(domains_hash))

        result = acme.download_order_certificate(order)

        logger.info("  Certificate downloaded.")
    except IOError as e:
        print("Connection or service request failed. Aborting.")
        raise AutomatoesError(e)

    try:
        certificates = strip_certificates(result.content)
        certificate = load_pem_certificate(certificates[0])

        # Print some neat info
        print("  Expires: {}".format(certificate.not_valid_after.strftime(
            EXPIRATION_FORMAT)))
        print("   SHA256: {}".format(binascii.hexlify(
            certificate.fingerprint(SHA256())).decode('ascii')))

        # Write the key, certificate and full chain
        os.makedirs(output_path, exist_ok=True)
        cert_path = os.path.join(output_path, domains[0] + '.crt')
        chain_path = os.path.join(output_path, domains[0] + '.chain.crt')
        intermediate_path = os.path.join(output_path,
                                         domains[0] + '.intermediate.crt')
        key_path = os.path.join(output_path, domains[0] + '.pem')

        if order.key is not None:
            with open(key_path, 'wb') as f:
                os.chmod(key_path, 0o600)
                f.write(order.key.encode('ascii'))
                print("\n  Wrote key to {}".format(f.name))

        with open(cert_path, 'wb') as f:
            f.write(export_pem_certificate(certificate))
            print("  Wrote certificate to {}".format(f.name))

        with open(chain_path, 'wb') as f:
            f.write(export_pem_certificate(certificate))
            if len(certificates) > 1:
                f.write(export_pem_certificate(load_pem_certificate(
                    certificates[1])))
            print("  Wrote certificate with intermediate to {}".format(f.name))

        if len(certificates) > 1:
            with open(intermediate_path, 'wb') as f:
                f.write(export_pem_certificate(load_pem_certificate(
                    certificates[1])))
                print("  Wrote intermediate certificate to {}".format(f.name))
    except IOError as e:
        print("  ERROR: Failed to write certificate or key. Going to print "
              "them for you instead.")
        if order.key is not None:
            for line in order.key.split('\n'):
                print("ERROR: {}".format(line))
        for line in export_pem_certificate(
                certificate).decode('ascii').split('\n'):
            print("ERROR: {}".format(line))
        raise AutomatoesError(e)



#!/usr/bin/env python
#
# Copyright 2019-2020 Flavio Garcia
# Copyright 2016-2017 Veeti Paananen under MIT License
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Cryptography, hopefully mostly correct.
"""

import base64
import json
import logging
# TODO: Remove that after python 3.5 depreciation
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from cryptography import x509
from cryptography.x509 import NameOID
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric.rsa import (
    generate_private_key,
    RSAPrivateKey,
)
from cryptography.hazmat.primitives.asymmetric.ec import (
    EllipticCurvePrivateKey,
)
from cryptography.hazmat.primitives.serialization import (
    load_pem_private_key,
    Encoding,
    PrivateFormat,
    NoEncryption,
)
from cryptography.hazmat.primitives import hashes

import re

logger = logging.getLogger(__name__)


def jose_b64(data):
    """
    Encodes data with JOSE/JWS base 64 encoding.
    """
    return base64.urlsafe_b64encode(data).decode('ascii').replace('=', '')


def generate_rsa_key(size=2048):
    """
    Generates a new RSA private key.
    """
    return generate_private_key(65537, size, default_backend())


def generate_header(account_key):
    """
    Creates a new request header for the specified account key.
    """
    numbers = account_key.public_key().public_numbers()
    e = numbers.e.to_bytes((numbers.e.bit_length() // 8 + 1), byteorder='big')
    n = numbers.n.to_bytes((numbers.n.bit_length() // 8 + 1), byteorder='big')
    if n[0] == 0: # for strict JWK
        n = n[1:]
    return {
        'alg': 'RS256',
        'jwk': {
            'kty': 'RSA',
            'e': jose_b64(e),
            'n': jose_b64(n),
        },
    }


def generate_jwk_thumbprint(account_key):
    """
    Generates a JWK thumbprint for the specified account key.
    """
    jwk = generate_header(account_key)['jwk']
    as_json = json.dumps(jwk, sort_keys=True, separators=(',', ':'))

    sha256 = hashes.Hash(hashes.SHA256(), default_backend())
    sha256.update(as_json.encode('utf-8'))

    return jose_b64(sha256.finalize())


def sign_request(key, header, protected_header, payload):
    """
    Creates a JSON Web Signature for the request header and payload using the
    specified account key.
    """
    protected = jose_b64(json.dumps(protected_header).encode('utf8'))
    payload = jose_b64(json.dumps(payload).encode('utf8'))
    data = "{protected}.{payload}".format(protected=protected, payload=payload)
    signed_data = key.sign(data.encode("ascii"), padding.PKCS1v15(),
                           hashes.SHA256())
    return json.dumps({
        'header': header,
        'protected': protected,
        'payload': payload,
        'signature': jose_b64(signed_data),
    })


def sign_request_v2(key, protected_header, payload):
    """
    Creates a JSON Web Signature for the request header and payload using the
    specified account key.
    """
    protected = jose_b64(json.dumps(protected_header).encode('utf8'))
    # Forced payload none for Post-as-Get
    if payload is not None and payload != "":
        payload = jose_b64(json.dumps(payload).encode('utf8'))
    elif payload is None:
        payload = ""
    data = "{protected}.{payload}".format(protected=protected, payload=payload)
    signed_data = key.sign(data.encode("ascii"), padding.PKCS1v15(),
                           hashes.SHA256())
    return json.dumps({
        'protected': protected,
        'payload': payload,
        'signature': jose_b64(signed_data),
    })


def load_private_key(data):
    """
    Loads a PEM-encoded private key.
    """
    key = load_pem_private_key(data, password=None, backend=default_backend())
    if not isinstance(key, (RSAPrivateKey, EllipticCurvePrivateKey)):
        raise ValueError("Key is not a private RSA or EC key.")
    elif isinstance(key, RSAPrivateKey) and key.key_size < 2048:
        raise ValueError("The key must be 2048 bits or longer.")

    return key


def export_private_key(key):
    """
    Exports a private key in OpenSSL PEM format.
    """
    return key.private_bytes(Encoding.PEM, PrivateFormat.TraditionalOpenSSL, NoEncryption())


def create_csr(key, domains, must_staple=False):
    """
    Creates a CSR in DER format for the specified key and domain names.
    """
    assert domains
    name = x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, domains[0]),
    ])
    san = x509.SubjectAlternativeName(
        [x509.DNSName(domain) for domain in domains])
    csr = (x509.CertificateSigningRequestBuilder()
           .subject_name(name).add_extension(san, critical=False))
    if must_staple:
        ocsp_must_staple = x509.TLSFeature(
            features=[x509.TLSFeatureType.status_request])
        csr = csr.add_extension(ocsp_must_staple, critical=False)
    return csr.sign(key, hashes.SHA256(), default_backend())


def export_csr_for_acme(csr):
    """
    Exports a X.509 CSR for the ACME protocol (JOSE Base64 DER).
    """
    return export_certificate_for_acme(csr)


def load_csr(data):
    """
    Loads a PEM X.509 CSR.
    """
    return x509.load_pem_x509_csr(data, default_backend())


def load_der_certificate(data):
    """
    Loads a DER X.509 certificate.
    """
    return x509.load_der_x509_certificate(data, default_backend())


def load_pem_certificate(data):
    """
    Loads a PEM X.509 certificate.
    """
    return x509.load_pem_x509_certificate(data, default_backend())


def get_certificate_domain_name(cert):
    return cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value


def get_certificate_domains(cert):
    """
    Gets a list of all Subject Alternative Names in the specified certificate.
    """
    for ext in cert.extensions:
        ext = ext.value
        if isinstance(ext, x509.SubjectAlternativeName):
            return ext.get_values_for_type(x509.DNSName)
    return []


def export_pem_certificate(cert):
    """
    Exports a X.509 certificate as PEM.
    """
    return cert.public_bytes(Encoding.PEM)


def export_certificate_for_acme(cert):
    """
    Exports a X.509 certificate for the ACME protocol (JOSE Base64 DER).
    """
    return jose_b64(cert.public_bytes(Encoding.DER))


def strip_certificates(data):
    p = re.compile("-----BEGIN CERTIFICATE-----\n(?s).+?"
                   "-----END CERTIFICATE-----\n")
    stripped_data = []
    for cert in p.findall(data.decode()):
        stripped_data.append(cert.encode())
    return stripped_data



# -*- coding: UTF-8 -*-
#
# Copyright 2019-2021 Flavio Garcia
# Copyright 2016-2017 Veeti Paananen under MIT License
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Flavio Garcia <piraz@candango.org>"
__version__ = (0, 9, 7)
__licence__ = "Apache License V2.0"


def get_version():
    return ".".join(map(str, __version__))


def get_author():
    return __author__.split(" <")[0]


def get_author_email():
    return __author__.split(" <")[1][:-1]



#!/usr/bin/env python
#
# Copyright 2019 Flavio Garcia
# Copyright 2016-2017 Veeti Paananen under MIT License
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class AutomatoesError(Exception):
    pass


class AcmeError(IOError):
    def __init__(self, response):
        message = "The ACME request failed."
        try:
            details = response.json()
            self.type = details.get('type', 'unknown')
            message = "{} (type {}, HTTP {})".format(details.get('detail'),
                    self.type, response.status_code)
        except (ValueError, TypeError, AttributeError):
            pass
        super().__init__(message)


class AccountAlreadyExistsError(AcmeError):

    def __init__(self, response, existing_uri):
        super().__init__(response)
        self.existing_uri = existing_uri



#!/usr/bin/env python
#
# Copyright 2019-2021 Flavio Garcia
# Copyright 2016-2017 Veeti Paananen under MIT License
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import automatoes
from codecs import open
from setuptools import setup
import sys

try:
    # for pip >= 10
    from pip._internal.req import parse_requirements
except ImportError:
    # for pip <= 9.0.3
    print("error: Upgrade to a pip version newer than 10. Run \"pip install "
          "--upgrade pip\".")
    sys.exit(1)

with open("README.md", "r") as fh:
    long_description = fh.read()


# Solution from http://bit.ly/29Yl8VN
def resolve_requires(requirements_file):
    try:
        requirements = parse_requirements("./%s" % requirements_file,
                                          session=False)
        return [str(ir.req) for ir in requirements]
    except AttributeError:
        # for pip >= 20.1.x
        # Need to run again as the first run was ruined by the exception
        requirements = parse_requirements("./%s" % requirements_file,
                                          session=False)
        # pr stands for parsed_requirement
        return [str(pr.requirement) for pr in requirements]


def use_right_cryptography(requirements):
    cryptography_req = "requirements/cryptography.txt"
    if sys.version_info.minor < 6:
        cryptography_req = "requirements/cryptography_legacy.txt"
    requirements.append(resolve_requires(cryptography_req)[0])
    return requirements


setup(
    name="automatoes",
    version=automatoes.get_version(),
    license=automatoes.__licence__,
    description=("Let's Encrypt/ACME V2 client replacement for Manuale. Manual"
                 " or automated your choice."),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/candango/automatoes",
    author=automatoes.get_author(),
    author_email=automatoes.get_author_email(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
    ],
    packages=["automatoes"],
    install_requires=use_right_cryptography(
        resolve_requires("requirements/basic.txt")
    ),
    entry_points={
        'console_scripts': [
            "automatoes = automatoes.cli:automatoes_main",
            "manuale = automatoes.cli:manuale_main",
        ],
    },
)
